@@ -1,6 +1,4 @@
-# vim: tabstop=4 shiftwidth=4 softtabstop=4
 # coding=utf8
-
 # Copyright 2014 IBM Corporation
 # Copyright 2015 Lenovo
 #
@@ -16,38 +14,77 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
-# This module provides access to SDR offered by a BMC
-# This data is common between 'sensors' and 'inventory' modules since SDR
-# is both used to enumerate sensors for sensor commands and FRU ids for FRU
-# commands
+"""This module provides access to SDR offered by a BMC
 
-# For now, we will not offer persistent SDR caching as we do in xCAT's IPMI
-# code.  Will see if it is adequate to advocate for high object reuse in a
-# persistent process for the moment.
+This data is common between 'sensors' and 'inventory' modules since SDR
+is both used to enumerate sensors for sensor commands and FRU ids for FRU
+commands
 
-# Focus is at least initially on the aspects that make the most sense for a
-# remote client to care about.  For example, smbus information is being
-# skipped for now
+For now, we will not offer persistent SDR caching as we do in xCAT's IPMI
+code.  Will see if it is adequate to advocate for high object reuse in a
+persistent process for the moment.
+
+Focus is at least initially on the aspects that make the most sense for a
+remote client to care about.  For example, smbus information is being
+skipped for now
+"""
 
 import math
+import os
+import random
+import string
+import struct
+import weakref
+
+import six
+
 import pyghmi.constants as const
 import pyghmi.exceptions as exc
 import pyghmi.ipmi.private.constants as ipmiconst
-import struct
-import weakref
+
+try:
+    import cPickle as pickle
+except ImportError:
+    import pickle
 
 TYPE_UNKNOWN = 0
 TYPE_SENSOR = 1
 TYPE_FRU = 2
 
+shared_sdrs = {}
+
+
+oem_type_offsets = {
+    343: {  # Intel
+        149: {  # Cascade Lake-AP
+            0x7a: {
+                0xda: {
+                    3: {
+                        'desc': 'Allowed',
+                        'severity': const.Health.Ok,
+                    },
+                    4: {
+                        'desc': 'Restricted',
+                        'severity': const.Health.Ok,
+                    },
+                    5: {
+                        'desc': 'Disabled',
+                        'severity': const.Health.Ok,
+                    },
+                },
+            },
+        },
+    },
+}
+
 
 def ones_complement(value, bits):
     # utility function to help with the large amount of 2s
     # complement prevalent in ipmi spec
     signbit = 0b1 << (bits - 1)
     if value & signbit:
-        #if negative, subtract 1, then take 1s
-        #complement given bits width
+        # if negative, subtract 1, then take 1s
+        # complement given bits width
         return 0 - (value ^ ((0b1 << bits) - 1))
     else:
         return value
@@ -58,8 +95,8 @@ def twos_complement(value, bits):
     # complement prevalent in ipmi spec
     signbit = 0b1 << (bits - 1)
     if value & signbit:
-        #if negative, subtract 1, then take 1s
-        #complement given bits width
+        # if negative, subtract 1, then take 1s
+        # complement given bits width
         return 0 - ((value - 1) ^ ((0b1 << bits) - 1))
     else:
         return value
@@ -193,6 +230,8 @@ class SensorReading(object):
         self.value = None
         self.imprecision = None
         self.states = []
+        self.state_ids = []
+        self.unavailable = 0
         try:
             self.health = reading['health']
             self.states = reading['states']
@@ -215,6 +254,7 @@ class SensorReading(object):
             'imprecision': self.imprecision,
             'name': self.name,
             'type': self.type,
+            'unavailable': self.unavailable,
             'health': self.health
         })
 
@@ -248,9 +288,13 @@ class SDREntry(object):
     external code to pay attention to this class.
     """
 
-    def __init__(self, entrybytes, ipmicmd, reportunsupported=False):
+    def __init__(self, entrybytes, ipmicmd, reportunsupported=False,
+                 mfg_id=0, prod_id=0):
+        self.mfg_id = mfg_id
+        self.prod_id = prod_id
         # ignore record id for now, we only care about the sensor number for
         # moment
+        self.readable = True
         self.reportunsupported = reportunsupported
         self.ipmicmd = ipmicmd
         if entrybytes[2] != 0x51:
@@ -258,12 +302,14 @@ class SDREntry(object):
             raise NotImplementedError
         self.rectype = entrybytes[3]
         self.linearization = None
-        #most important to get going are 1, 2, and 11
+        # most important to get going are 1, 2, and 11
         self.sdrtype = TYPE_SENSOR  # assume a sensor
         if self.rectype == 1:  # full sdr
             self.full_decode(entrybytes[5:])
         elif self.rectype == 2:  # full sdr
             self.compact_decode(entrybytes[5:])
+        elif self.rectype == 3:  # event only
+            self.eventonly_decode(entrybytes[5:])
         elif self.rectype == 8:  # entity association
             self.association_decode(entrybytes[5:])
         elif self.rectype == 0x11:  # FRU locator
@@ -309,9 +355,15 @@ class SDREntry(object):
 
     def association_decode(self, entry):
         # table 43-4 Entity Associaition Record
-        #TODO(jbjohnso): actually represent this data
+        # TODO(jbjohnso): actually represent this data
         self.sdrtype = TYPE_UNKNOWN
 
+    def eventonly_decode(self, entry):
+        # table 43-3 event_only sensor record
+        self._common_decode(entry)
+        self.sensor_name = self.tlv_decode(entry[11], entry[12:])
+        self.readable = False
+
     def compact_decode(self, entry):
         # table 43-2 compact sensor record
         self._common_decode(entry)
@@ -322,22 +374,30 @@ class SDREntry(object):
         return trapval + offset
 
     def _common_decode(self, entry):
-        # compact and full are very similar
+        # event only, compact and full are very similar
         # this function handles the common aspects of compact and full
         # offsets from spec, minus 6
+        self.sensor_lun = entry[1] & 0x03
         self.sensor_number = entry[2]
         self.entity = ipmiconst.entity_ids.get(
             entry[3], 'Unknown entity {0}'.format(entry[3]))
-        self.sensor_type_number = entry[7]
+        if self.rectype == 3:
+            self.sensor_type_number = entry[5]
+            self.reading_type = entry[6]  # table 42-1
+        else:
+            self.sensor_type_number = entry[7]
+            self.reading_type = entry[8]  # table 42-1
         try:
-            self.sensor_type = ipmiconst.sensor_type_codes[entry[7]]
+            self.sensor_type = ipmiconst.sensor_type_codes[
+                self.sensor_type_number]
         except KeyError:
-            self.sensor_type = "UNKNOWN type " + str(entry[7])
-        self.reading_type = entry[8]  # table 42-1
-            # 0: unspecified
-            # 1: generic threshold based
-            # 0x6f: discrete sensor-specific from table 42-3, sensor offsets
-            # all others per table 42-2, generic discrete
+            self.sensor_type = "UNKNOWN type " + str(self.sensor_type_number)
+        if self.rectype == 3:
+            return
+        # 0: unspecified
+        # 1: generic threshold based
+        # 0x6f: discrete sensor-specific from table 42-3, sensor offsets
+        # all others per table 42-2, generic discrete
         # numeric format is one of:
         # 0 - unsigned, 1 - 1s complement, 2 - 2s complement, 3 - ignore number
         # compact records are supposed to always write it as '3', presumably
@@ -362,14 +422,21 @@ class SDREntry(object):
         self.percent = ''
         if entry[15] & 1 == 1:
             self.percent = '% '
+        if self.sensor_type_number == 0xb:
+            if self.unit_mod == '':
+                if entry[16] == 6:
+                    self.sensor_type = 'Power'
+            elif self.unit_mod == ' ':
+                if entry[16] == 6 and entry[17] in (22, 23, 24):
+                    self.sensor_type = 'Energy'
         self.baseunit = unit_types[entry[16]]
         self.modunit = unit_types[entry[17]]
         self.unit_suffix = self.percent + self.baseunit + self.unit_mod + \
             self.modunit
 
     def full_decode(self, entry):
-        #offsets are table from spec, minus 6
-        #TODO(jbjohnso): table 43-13, put in constants to interpret entry[3]
+        # offsets are table from spec, minus 6
+        # TODO(jbjohnso): table 43-13, put in constants to interpret entry[3]
         self._common_decode(entry)
         # now must extract the formula data to transform values
         # entry[18 to entry[24].
@@ -396,6 +463,11 @@ class SDREntry(object):
                 mapping = ipmiconst.sensor_type_offsets
                 desc = mapping[self.sensor_type_number][state]['desc']
                 health = mapping[self.sensor_type_number][state]['severity']
+            elif self.reading_type >= 0x70 and self.reading_type <= 0x7f:
+                sensedata = oem_type_offsets[self.mfg_id][self.prod_id][
+                    self.reading_type][self.sensor_type_number][state]
+                desc = sensedata['desc']
+                health = sensedata['severity']
             else:
                 desc = "Unknown state %d" % state
                 health = const.Health.Warning
@@ -412,7 +484,7 @@ class SDREntry(object):
             'type': self.sensor_type,
             'id': self.sensor_number,
         }
-        if reading[1] & 0b100000:
+        if reading[1] & 0b100000 or not reading[1] & 0b1000000:
             output['unavailable'] = 1
             return SensorReading(output, self.unit_suffix)
         if self.numeric_format == 2:
@@ -501,8 +573,8 @@ class SDREntry(object):
             self._set_tmp_formula(value)
             linearization = 0
         # time to compute the pre-linearization value.
-        decoded = float((value * self.m + self.b) *
-                        (10 ** self.resultexponent))
+        decoded = float((value * self.m + self.b)
+                        * (10 ** self.resultexponent))
         if linearization == 0:
             return decoded
         elif linearization == 1:
@@ -526,21 +598,18 @@ class SDREntry(object):
         elif linearization == 10:
             return math.sqrt(decoded)
         elif linearization == 11:
-            return decoded ** (1.0/3)
+            return decoded ** (1.0 / 3)
         else:
             raise NotImplementedError
 
     def decode_formula(self, entry):
-        self.m = \
-            twos_complement(entry[0] + ((entry[1] & 0b11000000) << 2), 10)
+        self.m = twos_complement(entry[0] + ((entry[1] & 0b11000000) << 2), 10)
         self.tolerance = entry[1] & 0b111111
-        self.b = \
-            twos_complement(entry[2] + ((entry[3] & 0b11000000) << 2), 10)
-        self.accuracy = (entry[3] & 0b111111) + \
-            (entry[4] & 0b11110000) << 2
+        self.b = twos_complement(entry[2] + ((entry[3] & 0b11000000) << 2), 10)
+        self.accuracy = (entry[3] & 0b111111) + (entry[4] & 0b11110000) << 2
         self.accuracyexp = (entry[4] & 0b1100) >> 2
         self.direction = entry[4] & 0b11
-            #0 = n/a, 1 = input, 2 = output
+        # 0 = n/a, 1 = input, 2 = output
         self.resultexponent = twos_complement((entry[5] & 0b11110000) >> 4, 4)
         bexponent = twos_complement(entry[5] & 0b1111, 4)
         # might as well do the math to 'b' now rather than wait for later
@@ -553,7 +622,8 @@ class SDREntry(object):
             return ""
         if ipmitype == 0:  # Unicode per 43.15 in ipmi 2.0 spec
             # the spec is not specific about encoding, assuming utf8
-            return unicode(struct.pack("%dB" % len(data), *data), "utf_8")
+            return six.text_type(struct.pack("%dB" % len(data), *data),
+                                 "utf_8")
         elif ipmitype == 1:  # BCD '+'
             tmpl = "%02X" * len(data)
             tstr = tmpl % tuple(data)
@@ -565,14 +635,17 @@ class SDREntry(object):
             tstr = ""
             while len(data) >= 3:  # the packing only works with 3 byte chunks
                 tstr += chr((data[0] & 0b111111) + 0x20)
-                tstr += chr(((data[1] & 0b1111) << 2) +
-                            (data[0] >> 6) + 0x20)
-                tstr += chr(((data[2] & 0b11) << 4) +
-                            (data[1] >> 4) + 0x20)
+                tstr += chr(((data[1] & 0b1111) << 2) + (data[0] >> 6) + 0x20)
+                tstr += chr(((data[2] & 0b11) << 4) + (data[1] >> 4) + 0x20)
                 tstr += chr((data[2] >> 2) + 0x20)
+            if not isinstance(tstr, str):
+                tstr = tstr.decode('utf-8')
             return tstr
         elif ipmitype == 3:  # ACSII+LATIN1
-            return struct.pack("%dB" % len(data), *data)
+            ret = struct.pack("%dB" % len(data), *data)
+            if not isinstance(ret, str):
+                ret = ret.decode('utf-8')
+            return ret
 
 
 class SDR(object):
@@ -584,14 +657,15 @@ class SDR(object):
 
     :param ipmicmd: A Command class object
     """
-    def __init__(self, ipmicmd):
+    def __init__(self, ipmicmd, cachedir=None):
         self.ipmicmd = weakref.proxy(ipmicmd)
         self.sensors = {}
         self.fru = {}
+        self.cachedir = cachedir
         self.read_info()
 
     def read_info(self):
-        #first, we want to know the device id
+        # first, we want to know the device id
         rsp = self.ipmicmd.xraw_command(netfn=6, command=1)
         rsp['data'] = bytearray(rsp['data'])
         self.device_id = rsp['data'][0]
@@ -600,17 +674,25 @@ class SDR(object):
         # since that provides usefully distinct state and this does not
         self.fw_major = rsp['data'][2] & 0b1111111
         self.fw_minor = "%02X" % rsp['data'][3]  # BCD encoding, oddly enough
-        if rsp['data'][1] & 0b10000000:
-            # For lack of any system with 'device sdrs', raise an
-            # exception when they are encountered for now, implement or
-            # ignore later
-            raise NotImplementedError
         self.ipmiversion = rsp['data'][4]  # 51h = 1.5, 02h = 2.0
-        self.mfg_id = rsp['data'][8] << 16 + rsp['data'][7] << 8 + \
+        self.mfg_id = (rsp['data'][8] << 16) + (rsp['data'][7] << 8) + \
             rsp['data'][6]
-        self.prod_id = rsp['data'][10] << 8 + rsp['data'][9]
+        self.prod_id = (rsp['data'][10] << 8) + rsp['data'][9]
         if len(rsp['data']) > 11:
             self.aux_fw = self.decode_aux(rsp['data'][11:15])
+        if rsp['data'][1] & 0b10000000 and rsp['data'][5] & 0b10 == 0:
+            # The device has device sdrs, also does not support SDR repository
+            # device, so we are meant to use an alternative mechanism to get
+            # SDR data
+            if rsp['data'][5] & 1:
+                # The device has sensor device support, so in theory we should
+                # be able to proceed
+                # However at the moment, we haven't done so
+                raise NotImplementedError
+            return
+            # We have Device SDR, without SDR Repository device, but
+            # also without sensor device support, no idea how to
+            # continue
         self.get_sdr()
 
     def get_sdr_reservation(self):
@@ -626,22 +708,56 @@ class SDR(object):
             # we only understand SDR version 51h, the only version defined
             # at time of this writing
             raise NotImplementedError
-        #NOTE(jbjohnso): we actually don't need to care about 'numrecords'
+        # NOTE(jbjohnso): we actually don't need to care about 'numrecords'
         # since FFFF marks the end explicitly
-        #numrecords = (rsp['data'][2] << 8) + rsp['data'][1]
-        #NOTE(jbjohnso): don't care about 'free space' at the moment
-        #NOTE(jbjohnso): most recent timstamp data for add and erase could be
+        # numrecords = (rsp['data'][2] << 8) + rsp['data'][1]
+        # NOTE(jbjohnso): don't care about 'free space' at the moment
+        # NOTE(jbjohnso): most recent timstamp data for add and erase could be
         # handy to detect cache staleness, but for now will assume invariant
         # over life of session
-        #NOTE(jbjohnso): not looking to support the various options in op
+        # NOTE(jbjohnso): not looking to support the various options in op
         # support, ignore those for now, reservation if some BMCs can't read
         # full SDR in one slurp
+        modtime = struct.unpack('!Q', bytes(repinfo['data'][5:13]))[0]
         recid = 0
         rsvid = 0  # partial 'get sdr' will require this
         offset = 0
         size = 0xff
         chunksize = 128
+        try:
+            csdrs = shared_sdrs[
+                (self.fw_major, self.fw_minor, self.mfg_id, self.prod_id,
+                 self.device_id, modtime)]
+            self.sensors = csdrs['sensors']
+            self.fru = csdrs['fru']
+            return
+        except KeyError:
+            pass
+        cachefilename = None
         self.broken_sensor_ids = {}
+        if self.cachedir:
+            cachefilename = 'sdrcache.{0}.{1}.{2}.{3}.{4}.{5}'.format(
+                self.mfg_id, self.prod_id, self.device_id, self.fw_major,
+                self.fw_minor, modtime)
+            cachefilename = os.path.join(self.cachedir, cachefilename)
+        if cachefilename and os.path.isfile(cachefilename):
+            with open(cachefilename, 'rb') as cfile:
+                csdrs = pickle.load(cfile)
+                for sdrdata in csdrs:
+                    self.add_sdr(sdrdata)
+                for sid in self.broken_sensor_ids:
+                    try:
+                        del self.sensors[sid]
+                    except KeyError:
+                        pass
+                shared_sdrs[
+                    (self.fw_major, self.fw_minor, self.mfg_id, self.prod_id,
+                     self.device_id, modtime)] = {
+                    'sensors': self.sensors,
+                    'fru': self.fru,
+                }
+                return
+        sdrraw = [] if cachefilename else None
         while recid != 0xffff:  # per 33.12 Get SDR command, 0xffff marks end
             newrecid = 0
             currlen = 0
@@ -686,6 +802,8 @@ class SDR(object):
                 if (offset + size) > currlen:
                     size = currlen - offset
             self.add_sdr(sdrdata)
+            if sdrraw is not None:
+                sdrraw.append(sdrdata)
             offset = 0
             if size != 0xff:
                 size = 5
@@ -697,14 +815,28 @@ class SDR(object):
                 del self.sensors[sid]
             except KeyError:
                 pass
+        shared_sdrs[(self.fw_major, self.fw_minor, self.mfg_id, self.prod_id,
+                     self.device_id, modtime)] = {
+            'sensors': self.sensors,
+            'fru': self.fru,
+        }
+        if cachefilename:
+            suffix = ''.join(
+                random.choice(string.ascii_lowercase) for _ in range(12))
+            with open(cachefilename + '.' + suffix, 'wb') as cfile:
+                pickle.dump(sdrraw, cfile)
+            os.rename(cachefilename + '.' + suffix, cachefilename)
 
     def get_sensor_numbers(self):
-        return self.sensors.iterkeys()
+        for number in self.sensors:
+            if self.sensors[number].readable:
+                yield number
 
     def add_sdr(self, sdrbytes):
-        newent = SDREntry(sdrbytes, self.ipmicmd)
+        newent = SDREntry(sdrbytes, self.ipmicmd, False, self.mfg_id,
+                          self.prod_id)
         if newent.sdrtype == TYPE_SENSOR:
-            id = newent.sensor_number
+            id = '{0}.{1}'.format(newent.sensor_number, newent.sensor_lun)
             if id in self.sensors:
                 self.broken_sensor_ids[id] = True
                 return
@@ -720,20 +852,3 @@ class SDR(object):
         # This is where manufacturers can add their own
         # decode information
         return "".join(hex(x) for x in auxdata)
-
-if __name__ == "__main__":  # test code
-    import os
-    import pyghmi.ipmi.command as ipmicmd
-    import sys
-    password = os.environ['IPMIPASSWORD']
-    bmc = sys.argv[1]
-    user = sys.argv[2]
-    ipmicmd = ipmicmd.Command(bmc=bmc, userid=user, password=password)
-    sdr = SDR(ipmicmd)
-    for number in sdr.get_sensor_numbers():
-        rsp = ipmicmd.raw_command(command=0x2d, netfn=4, data=(number,))
-        if 'error' in rsp:
-            continue
-        reading = sdr.sensors[number].decode_sensor_reading(rsp['data'])
-        if reading is not None:
-            print(repr(reading))
